"""Dataset Tests (core methods)"""
import copy
import json
import re
import tempfile
from os import remove
from os.path import join

import pytest
from hdx.utilities.dictandlist import merge_two_dictionaries
from hdx.utilities.loader import load_yaml

from hdx.api import __version__
from hdx.api.configuration import Configuration
from hdx.data.dataset import Dataset, NotRequestableError
from hdx.data.hdxobject import HDXError
from hdx.data.resource import Resource
from hdx.data.resource_view import ResourceView

from . import (
    MockResponse,
    dataset_data,
    dataset_mockshow,
    dataset_resultdict,
    resources_data,
)
from .test_resource_view import (
    resource_view_list,
    resource_view_mockcreate,
    resource_view_mocklist,
    resource_view_mockshow,
)
from .test_vocabulary import vocabulary_mockshow

searchdict = load_yaml(join("tests", "fixtures", "dataset_search_results.yml"))
dataset_list = [
    "acled-conflict-data-for-libya",
    "acled-conflict-data-for-liberia",
    "acled-conflict-data-for-lesotho",
    "acled-conflict-data-for-kenya",
    "acled-conflict-data-for-guinea",
    "acled-conflict-data-for-ghana",
    "acled-conflict-data-for-gambia",
    "acled-conflict-data-for-gabon",
    "acled-conflict-data-for-ethiopia",
    "acled-conflict-data-for-eritrea",
]
hxlupdate_list = [
    {
        "title": "Quick Charts",
        "resource_id": "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
        "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
        "view_type": "hdx_hxl_preview",
        "description": "",
        "id": "29cc5894-4306-4bef-96ce-b7a833e7986a",
    }
]
dataset_autocomplete = [
    {
        "match_field": "name",
        "match_displayed": "acled-conflict-data-for-africa-1997-lastyear",
        "name": "acled-conflict-data-for-africa-1997-lastyear",
        "title": "ACLED Conflict Data for Africa 1997-2016",
    }
]


def mocksearch(url, datadict):
    if "search" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not search", "__type": "TEST ERROR: Not Search Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
        )
    if datadict["q"] == "ACLED":
        newsearchdict = copy.deepcopy(searchdict)
        if datadict["rows"] == 11:
            newsearchdict["results"].append(newsearchdict["results"][0])
        elif datadict["rows"] == 6:
            if datadict["start"] == 2:
                newsearchdict["results"] = newsearchdict["results"][2:8]
        elif datadict["rows"] == 5:
            if datadict["start"] == 0:
                newsearchdict["count"] = 5
                newsearchdict["results"] = newsearchdict["results"][:5]
            elif datadict["start"] == 5:
                newsearchdict["count"] = 6  # return wrong count
                newsearchdict["results"] = newsearchdict["results"][:5]
            else:
                newsearchdict["count"] = 0
                newsearchdict["results"] = list()
        result = json.dumps(newsearchdict)
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}'
            % result,
        )
    if datadict["q"] == '"':
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "Validation Error", "__type": "Validation Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
        )
    if datadict["q"] == "ajyhgr":
        return MockResponse(
            200,
            '{"success": true, "result": {"count": 0, "results": []}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
        )
    return MockResponse(
        404,
        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
    )


def mocklist(url, datadict):
    if "list" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not all", "__type": "TEST ERROR: Not All Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_list"}',
        )

    offset = datadict.get("offset", 0)
    limit = datadict.get("limit", len(dataset_list))
    result = json.dumps(dataset_list[offset : offset + limit])
    return MockResponse(
        200,
        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_list"}'
        % result,
    )


def mockall(url, datadict):
    if "search" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not search", "__type": "TEST ERROR: Not All Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
        )
    newsearchdict = copy.deepcopy(searchdict)
    if datadict["rows"] == 11:
        newsearchdict["results"].append(newsearchdict["results"][0])
    elif datadict["rows"] == 7:
        if datadict["start"] == 2:
            newsearchdict["results"] = newsearchdict["results"][2:9]
        else:
            newsearchdict["results"] = newsearchdict["results"][
                4:5
            ]  # repeated dataset
    elif datadict["rows"] == 5:
        newsearchdict["count"] = 6
        if datadict["sort"] == "metadata_modified desc":
            if datadict["start"] == 0:
                newsearchdict["results"] = newsearchdict["results"][:5]
            else:
                newsearchdict["results"] = newsearchdict["results"][
                    4:5
                ]  # repeated dataset
        elif datadict["sort"] == "metadata_created asc":
            if datadict["start"] == 0:
                newsearchdict["results"] = newsearchdict["results"][:5]
            else:
                newsearchdict["results"] = newsearchdict["results"][5:6]
        else:
            if datadict["start"] == 0:
                newsearchdict["results"] = newsearchdict["results"][:5]
            else:
                newsearchdict["count"] = 7
                newsearchdict["results"] = newsearchdict["results"][5:7]
    result = json.dumps(newsearchdict)
    return MockResponse(
        200,
        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}'
        % result,
    )


def mockhxlupdate(url, datadict):
    if "hxl" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not HXL Update", "__type": "TEST ERROR: Not HXL Update Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_hxl_update"}',
        )
    result = json.dumps(hxlupdate_list)
    return MockResponse(
        200,
        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_hxl_update"}'
        % result,
    )


class TestDatasetCore:
    @pytest.fixture(scope="class")
    def test_file(self):
        return join("tests", "fixtures", "test_data.csv")

    @pytest.fixture(scope="class")
    def static_yaml(self):
        return join("tests", "fixtures", "config", "hdx_dataset_static.yml")

    @pytest.fixture(scope="class")
    def static_json(self):
        return join("tests", "fixtures", "config", "hdx_dataset_static.json")

    @pytest.fixture(scope="function")
    def read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return dataset_mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_revise(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                if isinstance(data, dict):
                    datadict = {
                        k.decode("utf8"): v.decode("utf8")
                        for k, v in data.items()
                    }
                else:
                    datadict = json.loads(data.decode("utf-8"))
                if datadict["match"] == '{"name":"MyDataset1"}':
                    resultdictcopy = copy.deepcopy(dataset_resultdict)
                    resultdictcopy["resources"][0]["description"] = "haha"
                    resultdictcopy = {"package": resultdictcopy}
                    result = json.dumps(resultdictcopy)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_revise"}'
                        % result,
                    )
                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_revise"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_create(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                if isinstance(data, dict):
                    datadict = {
                        k.decode("utf8"): v.decode("utf8")
                        for k, v in data.items()
                    }
                else:
                    datadict = json.loads(data.decode("utf-8"))
                if "vocabulary" in url:
                    return vocabulary_mockshow(url, datadict)
                if "show" in url:
                    return dataset_mockshow(url, datadict)
                if "hxl" in url:
                    return mockhxlupdate(url, datadict)
                if "default" in url:
                    result = json.dumps(resource_view_list)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_create_default_resource_views"}'
                        % result,
                    )
                if "resource_view" in url:
                    result = json.dumps(resource_view_list[1])
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_create"}'
                        % result,
                    )
                if "resource" in url:
                    result = json.dumps(resources_data[0])
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_create"}'
                        % result,
                    )
                if "create" not in url and "revise" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not create", "__type": "TEST ERROR: Not Create Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_create"}',
                    )
                if "revise" in url:
                    datadict = json.loads(datadict["update"])
                if datadict["name"] == "MyDataset1":
                    resultdictcopy = copy.deepcopy(dataset_resultdict)
                    resultdictcopy["state"] = datadict["state"]
                    if "revise" in url:
                        resultdictcopy = {"package": resultdictcopy}
                    result = json.dumps(resultdictcopy)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_create"}'
                        % result,
                    )
                if datadict["name"] == "MyDataset2":
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_create"}',
                    )
                if datadict["name"] == "MyDataset3":
                    return MockResponse(
                        200,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_create"}',
                    )
                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_create"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_update(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                if isinstance(data, dict):
                    datadict = {
                        k.decode("utf8"): v.decode("utf8")
                        for k, v in data.items()
                    }
                else:
                    datadict = json.loads(data.decode("utf-8"))
                if "default" in url:
                    result = json.dumps(resource_view_list)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_create_default_resource_views"}'
                        % result,
                    )
                if "resource_view" in url:
                    if "show" in url:
                        return resource_view_mockshow(url, datadict)
                    if "list" in url:
                        return resource_view_mocklist(url, datadict)
                    if "create" in url:
                        if datadict["title"] == "Quick Charts":
                            return resource_view_mockcreate(url, datadict)
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not create", "__type": "TEST ERROR: Not Create Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_create"}',
                    )
                if "vocabulary" in url:
                    return vocabulary_mockshow(url, datadict)
                if "show" in url:
                    return dataset_mockshow(url, datadict)
                if "hxl" in url:
                    return mockhxlupdate(url, datadict)
                if "resource" in url:
                    resultdictcopy = copy.deepcopy(resources_data[0])
                    merge_two_dictionaries(resultdictcopy, datadict)
                    result = json.dumps(resultdictcopy)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_update"}'
                        % result,
                    )
                else:
                    if "revise" not in url:
                        return MockResponse(
                            404,
                            '{"success": false, "error": {"message": "TEST ERROR: Not update", "__type": "TEST ERROR: Not Update Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_update"}',
                        )
                    datadict = json.loads(datadict["update"])
                    if datadict["name"] in ["MyDataset1", "DatasetExist"]:
                        resultdictcopy = copy.deepcopy(dataset_resultdict)
                        merge_two_dictionaries(resultdictcopy, datadict)
                        for i, resource in enumerate(
                            resultdictcopy["resources"]
                        ):
                            for j, resource2 in enumerate(
                                resultdictcopy["resources"]
                            ):
                                if i != j:
                                    if resource == resource2:
                                        del resultdictcopy["resources"][j]
                                        break
                            resource["package_id"] = resultdictcopy["id"]
                        resultdictcopy = {"package": resultdictcopy}
                        result = json.dumps(resultdictcopy)
                        return MockResponse(
                            200,
                            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_update"}'
                            % result,
                        )
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_update"}',
                    )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_reorder(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "show" in url:
                    return dataset_mockshow(url, datadict)
                if "hxl" in url:
                    return mockhxlupdate(url, datadict)
                if "reorder" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not reorder", "__type": "TEST ERROR: Not Reorder Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_resource_reorder"}',
                    )
                if datadict["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_resource_reorder"}'
                        % decodedata,
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_resource_reorder"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_delete(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "show" in url:
                    return dataset_mockshow(url, datadict)
                if "purge" not in url and "delete" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not delete", "__type": "TEST ERROR: Not Delete Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_delete"}',
                    )
                if "resource" in url:
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_delete"}'
                        % decodedata,
                    )

                if datadict["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_delete"}'
                        % decodedata,
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=dataset_delete"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def search(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return mocksearch(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_list(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return mocklist(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def all(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return mockall(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_autocomplete(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "autocomplete" not in url or "acled" not in datadict["q"]:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not autocomplete", "__type": "TEST ERROR: Not Autocomplete Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_autocomplete"}',
                    )
                result = json.dumps(dataset_autocomplete)
                return MockResponse(
                    200,
                    '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_autocomplete"}'
                    % result,
                )

        Configuration.read().remoteckan().session = MockSession()

    def test_read_from_hdx(self, configuration, read):
        dataset = Dataset.read_from_hdx("TEST1")
        assert dataset["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        assert dataset["name"] == "MyDataset1"
        assert dataset["dataset_date"] == "06/04/2016"
        assert dataset.number_of_resources() == 3
        dataset = Dataset.read_from_hdx("TEST2")
        assert dataset is None
        dataset = Dataset.read_from_hdx("TEST3")
        assert dataset is None

    def test_revise(self, configuration, test_file, post_revise):
        dataset = Dataset.revise(
            {"name": "MyDataset1"},
            filter=["-resources__0__description"],
            update={"resources": [{"description": "haha"}]},
            files_to_upload={"update__resources__0__upload": test_file},
        )
        assert dataset["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        assert dataset["name"] == "MyDataset1"
        assert dataset["dataset_date"] == "06/04/2016"
        assert dataset.get_resource()["description"] == "haha"
        with pytest.raises(HDXError):
            Dataset.revise(
                {"name": "MyDataset1"},
                filter=["-resources__0__description"],
                update={"resources": [{"description": "haha"}]},
                files_to_upload={"update__resources__0__upload": "NOTEXIST"},
            )
        with pytest.raises(HDXError):
            dataset._write_to_hdx(
                "revise",
                dict(),
                id_field_name="",
                files_to_upload={"update__resources__0__upload": "NOTEXIST"},
            )

    def test_create_in_hdx(self, configuration, post_create):
        dataset = Dataset()
        with pytest.raises(HDXError):
            dataset.create_in_hdx()
        dataset["id"] = "TEST1"
        dataset["name"] = "LALA"
        with pytest.raises(HDXError):
            dataset.create_in_hdx()

        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        with pytest.raises(HDXError):
            dataset.create_in_hdx()
        dataset.create_in_hdx(allow_no_resources=True)
        assert dataset["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        del resource["description"]
        del dataset["tags"]
        dataset.add_update_resources([resource, resource])
        with pytest.raises(HDXError):
            dataset.create_in_hdx()
        dataset.create_in_hdx(ignore_fields=["tags", "resource:description"])
        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        dataset.add_update_resources([resource, resource])
        dataset.create_in_hdx()
        assert dataset["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 3

        datasetdata["name"] = "MyDataset2"
        dataset = Dataset(datasetdata)
        with pytest.raises(HDXError):
            dataset.create_in_hdx()

        datasetdata["name"] = "MyDataset3"
        dataset = Dataset(datasetdata)
        with pytest.raises(HDXError):
            dataset.create_in_hdx()

        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        dataset.add_update_resource(resource)
        del dataset["tags"]
        with pytest.raises(HDXError):
            dataset.create_in_hdx()
        dataset.create_in_hdx(ignore_check=True)

        config = Configuration(user_agent="test", hdx_read_only=True)
        config.setup_session_remoteckan()
        config.remoteckan().session = Configuration.read().remoteckan().session
        uniqueval = "myconfig"
        config.unique = uniqueval

        datasetdata = copy.deepcopy(dataset_data)
        resourcesdata = copy.deepcopy(resources_data)
        datasetdata["resources"] = resourcesdata
        dataset = Dataset(datasetdata)
        assert len(dataset.get_resources()) == 3
        del datasetdata["resources"]
        uniqueval = "myconfig2"
        config.unique = uniqueval
        dataset = Dataset(datasetdata, configuration=config)
        del resourcesdata[0]["id"]
        del resourcesdata[1]["id"]
        dataset.add_update_resources(resourcesdata)
        dataset.create_in_hdx()
        assert dataset["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        assert len(dataset.resources) == 3
        assert dataset.resources[0].configuration.unique == uniqueval
        assert dataset["state"] == "active"
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        resource = Resource(resourcesdata[0])
        file = tempfile.NamedTemporaryFile(delete=False)
        resource.set_file_to_upload(file.name)
        dataset.add_update_resource(resource)
        dataset.create_in_hdx()
        remove(file.name)
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 3
        # Dataset creates that end up updating are in the test below

    def test_update_in_hdx(self, configuration, post_update):
        dataset = Dataset()
        dataset["id"] = "NOTEXIST"
        with pytest.raises(HDXError):
            dataset.update_in_hdx()
        dataset["name"] = "LALA"
        with pytest.raises(HDXError):
            dataset.update_in_hdx()

        dataset = Dataset.read_from_hdx("TEST1")
        assert dataset["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        assert dataset["dataset_date"] == "06/04/2016"

        dataset["dataset_date"] = "02/26/2016"
        dataset["id"] = "TEST1"
        dataset["name"] = "MyDataset1"
        dataset.update_in_hdx()
        assert dataset["id"] == "TEST1"
        assert dataset["dataset_date"] == "02/26/2016"
        assert dataset["state"] == "active"
        pattern = (
            r"HDXPythonLibrary/%s-test \([12]\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\d.\d\d\d\d\d\d\)"
            % __version__
        )
        match = re.search(pattern, dataset["updated_by_script"])
        assert match
        resourceviewdata = {
            "title": "Quick Charts",
            "resource_name": "Resource1",
            "view_type": "hdx_hxl_preview",
        }
        dataset.preview_resourceview = ResourceView(resourceviewdata)
        dataset.update_in_hdx()
        assert dataset.preview_resourceview is None
        dataset.update_in_hdx(updated_by_script="hi")
        pattern = r"hi \([12]\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\d.\d\d\d\d\d\d\)"
        match = re.search(pattern, dataset["updated_by_script"])
        assert match
        dataset.update_in_hdx(
            updated_by_script="hi",
            batch="6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
        )
        match = re.search(pattern, dataset["updated_by_script"])
        assert match
        assert dataset["batch"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        assert dataset["batch_mode"] == "DONT_GROUP"
        with pytest.raises(HDXError):
            dataset.update_in_hdx(updated_by_script="hi", batch="1234")

        dataset["id"] = "NOTEXIST"
        with pytest.raises(HDXError):
            dataset.update_in_hdx()

        del dataset["id"]
        with pytest.raises(HDXError):
            dataset.update_in_hdx()

        dataset["id"] = "TEST1"
        dataset["groups"] = list()
        with pytest.raises(HDXError):
            dataset.update_in_hdx()
        dataset.update_in_hdx(ignore_check=True)

        # These dataset creates actually do updates
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        dataset["id"] = "NOTEXIST"
        dataset["name"] = "DatasetExist"
        dataset.create_in_hdx(allow_no_resources=True)
        assert dataset["name"] == "DatasetExist"
        datasetdata["name"] = "MyDataset1"
        datasetdata["id"] = "TEST1"
        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        dataset.add_update_resources([resource, resource])
        dataset.create_in_hdx()
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 3

        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        dataset.add_update_resources([resource, resource])
        dataset.create_in_hdx(match_resources_by_metadata=False)
        assert dataset["id"] == "TEST1"
        assert dataset["dataset_date"] == "06/04/2016"
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 3
        dataset.update_in_hdx()
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 3
        dataset = Dataset.read_from_hdx("TEST4")
        dataset["id"] = "TEST4"
        assert dataset["state"] == "active"
        dataset.update_in_hdx()
        assert len(dataset.resources) == 3
        dataset = Dataset.read_from_hdx("TEST4")
        file = tempfile.NamedTemporaryFile(delete=False)
        resource.set_file_to_upload(file.name)
        dataset.add_update_resource(resource)
        dataset.update_in_hdx(batch="6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d")
        assert len(dataset.resources) == 3
        resource["name"] = "123"
        resource.set_file_to_upload(None)
        resource["url"] = "http://lala"
        dataset.add_update_resource(resource)
        dataset.update_in_hdx()
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 3
        del resource["id"]
        resource["name"] = "Resource1"
        dataset.add_update_resource(resource)
        dataset.update_in_hdx()
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 4

        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        dataset.add_update_resource(resource)
        resource = Resource(resourcesdata[1])
        dataset.add_update_resource(resource)
        resource = Resource(resourcesdata[0])
        resource["name"] = "ResourcePosition"
        resource["id"] = "123"
        resource.set_file_to_upload(file.name)
        dataset.add_update_resource(resource)
        resource = Resource(resourcesdata[0])
        resource["name"] = "changed name"
        resource["id"] = "456"
        resource.set_file_to_upload(file.name)
        dataset.update_in_hdx(match_resources_by_metadata=True)
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 4

        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        dataset.add_update_resource(resource)
        resource = Resource(resourcesdata[1])
        dataset.add_update_resource(resource)
        resource = Resource(resourcesdata[0])
        resource["name"] = "ResourcePosition"
        resource["id"] = "123"
        resource.set_file_to_upload(file.name)
        dataset.add_update_resource(resource)
        resource = dataset.get_resources()[0]
        resource["name"] = "changed name"
        resource["id"] = "456"
        resource.set_file_to_upload(file.name)
        dataset.update_in_hdx(match_resources_by_metadata=False)
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 3
        remove(file.name)
        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        dataset.add_update_resource(resource)
        dataset.update_in_hdx(remove_additional_resources=False)
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 3
        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        dataset.add_update_resource(resource)
        dataset.update_in_hdx(remove_additional_resources=True)
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 1
        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        dataset.add_update_resource(resource)
        dataset.update_in_hdx(
            match_resources_by_metadata=False, remove_additional_resources=True
        )
        assert dataset["state"] == "active"
        assert len(dataset.resources) == 1

        dataset = Dataset.read_from_hdx("TEST1")
        resources = dataset.get_resources()
        resource_ids = [x["id"] for x in resources]
        assert resource_ids == [
            "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
            "3d777226-96aa-4239-860a-703389d16d1f",
            "3d777226-96aa-4239-860a-703389d16d1g",
        ]
        dataset["id"] = "TEST1"
        dataset["name"] = "MyDataset1"
        dataset.resources = [resources[2], resources[0], resources[1]]
        dataset.update_in_hdx(match_resource_order=True)
        resource_ids = [x["id"] for x in dataset.get_resources()]
        assert resource_ids == [
            "3d777226-96aa-4239-860a-703389d16d1g",
            "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
            "3d777226-96aa-4239-860a-703389d16d1f",
        ]

    def test_delete_from_hdx(self, configuration, post_delete):
        dataset = Dataset.read_from_hdx("TEST1")
        dataset.delete_from_hdx()
        del dataset["id"]
        with pytest.raises(HDXError):
            dataset.delete_from_hdx()

    def test_update_yaml(self, configuration, static_yaml):
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        assert dataset["name"] == "MyDataset1"
        assert dataset["author"] == "AN Other"
        dataset.update_from_yaml(static_yaml)
        assert dataset["name"] == "MyDataset1"
        assert dataset["author"] == "acled"
        assert dataset.get_resources() == [
            {
                "id": "ABC",
                "description": "Resource1",
                "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
                "name": "Resource1",
                "url": "http://resource1.xlsx",
                "format": "xlsx",
            },
            {
                "id": "DEF",
                "description": "Resource2",
                "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
                "name": "Resource2",
                "url": "http://resource2.csv",
                "format": "csv",
            },
        ]
        dataset.get_resources()[0]["url"] = "http://lalala.xlsx"
        assert dataset.get_resources() == [
            {
                "id": "ABC",
                "description": "Resource1",
                "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
                "name": "Resource1",
                "url": "http://lalala.xlsx",
                "format": "xlsx",
            },
            {
                "id": "DEF",
                "description": "Resource2",
                "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
                "name": "Resource2",
                "url": "http://resource2.csv",
                "format": "csv",
            },
        ]
        dataset.update_from_yaml(static_yaml)
        assert dataset.get_resources() == [
            {
                "id": "ABC",
                "description": "Resource1",
                "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
                "name": "Resource1",
                "url": "http://resource1.xlsx",
                "format": "xlsx",
            },
            {
                "id": "DEF",
                "description": "Resource2",
                "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
                "name": "Resource2",
                "url": "http://resource2.csv",
                "format": "csv",
            },
        ]

    def test_update_json(self, configuration, static_json):
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        assert dataset["name"] == "MyDataset1"
        assert dataset["author"] == "AN Other"
        dataset.update_from_json(static_json)
        assert dataset["name"] == "MyDataset1"
        assert dataset["author"] == "Someone"
        assert dataset.get_resource() == {
            "id": "123",
            "description": "Resource1",
            "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
            "name": "Resource1",
            "url": "http://resource1.xlsx",
            "format": "xlsx",
        }
        assert dataset.get_resources() == [
            {
                "id": "123",
                "description": "Resource1",
                "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
                "name": "Resource1",
                "url": "http://resource1.xlsx",
                "format": "xlsx",
            },
            {
                "id": "456",
                "description": "Resource2",
                "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
                "name": "Resource2",
                "url": "http://resource2.csv",
                "format": "csv",
            },
        ]

    def test_add_update_delete_resources(self, configuration, post_delete):
        datasetdata = copy.deepcopy(dataset_data)
        resourcesdata = copy.deepcopy(resources_data)
        dataset = Dataset(datasetdata)
        dataset.add_update_resources(resourcesdata)
        dataset.add_update_resources(resourcesdata)
        assert len(dataset.resources) == 3
        dataset.delete_resource("de6549d8-268b-4dfe-adaf-a4ae5c8510d6")
        assert len(dataset.resources) == 3
        dataset.delete_resource("de6549d8-268b-4dfe-adaf-a4ae5c8510d5")
        assert len(dataset.resources) == 2
        resourcesdata = copy.deepcopy(resources_data)
        resource = Resource(resourcesdata[0])
        resource.set_file_to_upload("lala")
        dataset.add_update_resource(resource)
        assert dataset.resources[2].get_file_to_upload() == "lala"
        dataset.add_update_resource("de6549d8-268b-4dfe-adaf-a4ae5c8510d5")
        assert len(dataset.resources) == 3
        with pytest.raises(HDXError):
            dataset.add_update_resource(123)
        with pytest.raises(HDXError):
            dataset.add_update_resource("123")
        resourcesdata[0]["package_id"] = "123"
        with pytest.raises(HDXError):
            dataset.add_update_resources(resourcesdata)
        with pytest.raises(HDXError):
            dataset.add_update_resources(123)
        with pytest.raises(HDXError):
            dataset.delete_resource("NOTEXIST")
        datasetdata["resources"] = resourcesdata
        dataset = Dataset(datasetdata)
        assert len(dataset.resources) == 3
        resource = copy.deepcopy(resources_data[0])
        del resource["id"]
        dataset.add_update_resource(resource)
        assert len(dataset.resources) == 3
        dataset.add_update_resources([resource])
        assert len(dataset.resources) == 3
        resource = copy.deepcopy(resources_data[2])
        del resource["id"]
        dataset.add_update_resource(resource)
        assert len(dataset.resources) == 3
        dataset.add_update_resources([resource])
        assert len(dataset.resources) == 3
        resource["format"] = "mdb"
        dataset.add_update_resource(resource)
        assert len(dataset.resources) == 4
        resource = copy.deepcopy(resources_data[2])
        del resource["id"]
        resource["format"] = "doc"
        dataset.add_update_resources([resource])
        assert len(dataset.resources) == 5
        resource["format"] = "NOTEXIST"
        with pytest.raises(HDXError):
            dataset.add_update_resource(resource)

    def test_reorder_resources(self, configuration, post_reorder):
        dataset = Dataset.read_from_hdx("TEST1")
        resource_ids = [x["id"] for x in dataset.get_resources()]
        assert resource_ids == [
            "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
            "3d777226-96aa-4239-860a-703389d16d1f",
            "3d777226-96aa-4239-860a-703389d16d1g",
        ]
        dataset.reorder_resources(
            [
                "3d777226-96aa-4239-860a-703389d16d1f",
                "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
                "3d777226-96aa-4239-860a-703389d16d1g",
            ]
        )
        resource_ids = [x["id"] for x in dataset.get_resources()]
        assert resource_ids == [
            "3d777226-96aa-4239-860a-703389d16d1f",
            "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
            "3d777226-96aa-4239-860a-703389d16d1g",
        ]
        del dataset["id"]
        with pytest.raises(HDXError):
            dataset.reorder_resources(
                [
                    "3d777226-96aa-4239-860a-703389d16d1f",
                    "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
                    "3d777226-96aa-4239-860a-703389d16d1g",
                ]
            )

    def test_search_in_hdx(self, configuration, search):
        datasets = Dataset.search_in_hdx("ACLED")
        assert len(datasets) == 10
        datasets = Dataset.search_in_hdx("ACLED", offset=2, limit=6)
        assert len(datasets) == 6
        datasets = Dataset.search_in_hdx("ajyhgr")
        assert len(datasets) == 0
        with pytest.raises(HDXError):
            Dataset.search_in_hdx('"')
        with pytest.raises(HDXError):
            Dataset.search_in_hdx("ACLED", rows=11)
        with pytest.raises(HDXError):
            # Test returned row counts per page mismatch (wrong count of 6 purposely in mocksearch)
            Dataset.search_in_hdx("ACLED", page_size=5)

    def test_get_all_dataset_names(self, configuration, post_list):
        dataset_names = Dataset.get_all_dataset_names()
        assert dataset_names == dataset_list
        dataset_names = Dataset.get_all_dataset_names(start=3, rows=5)
        assert dataset_names == dataset_list[3:8]

    def test_get_all_datasets(self, configuration, all):
        datasets = Dataset.get_all_datasets()
        assert len(datasets) == 10
        datasets = Dataset.get_all_datasets(start=2, rows=7)
        assert len(datasets) == 7
        with pytest.raises(HDXError):
            Dataset.get_all_datasets(limit=11)
        with pytest.raises(HDXError):
            Dataset.get_all_datasets(
                page_size=5, sort="metadata_modified desc"
            )  # test repeated dataset (see mockall)
        with pytest.raises(HDXError):
            Dataset.get_all_datasets(
                page_size=5, sort="relevance desc"
            )  # test changed count (see mockall)
        datasets = Dataset.get_all_datasets(
            page_size=5
        )  # test no repeating dataset (see mockall)
        assert len(datasets) == 6

    def test_get_all_resources(self, configuration, search):
        datasets = Dataset.search_in_hdx("ACLED")
        resources = Dataset.get_all_resources(datasets)
        assert len(resources) == 3

    def test_hdxconnect(self, configuration, post_create):
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        dataset["private"] = True
        dataset.set_requestable()
        assert dataset["is_requestdata_type"] is True
        assert dataset.is_requestable() is True
        assert dataset["private"] is False
        dataset["private"] = True
        dataset.set_requestable(False)
        assert dataset["is_requestdata_type"] is False
        assert dataset["private"] is True
        dataset.set_requestable()
        assert dataset.get("field_name") is None
        assert dataset.get_fieldnames() == list()
        assert dataset.add_fieldname("myfield1") is True
        assert dataset.add_fieldnames(["myfield1", "myfield2"]) is False
        assert dataset.remove_fieldname("myfield1") is True
        assert dataset.remove_fieldname("myfield1") is False
        assert dataset.add_fieldnames(["myfield3", "myfield4"]) is True
        assert dataset.get_fieldnames() == ["myfield2", "myfield3", "myfield4"]
        assert dataset.get("fiele_types") is None
        assert dataset.get_filetypes() == list()
        assert dataset.add_filetype("mytype1") is True
        assert dataset.add_filetypes(["mytype1", "mytype2"]) is False
        assert dataset.remove_filetype("mytype1") is True
        assert dataset.remove_filetype("mytype1") is False
        assert dataset.add_filetypes(["mytype3", "mytype4"]) is True
        assert dataset.get_filetypes() == ["mytype2", "mytype3", "mytype4"]
        with pytest.raises(HDXError):
            dataset.create_in_hdx()
        dataset["num_of_rows"] = 100
        dataset.create_in_hdx()
        assert dataset["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        resourcesdata = copy.deepcopy(resources_data)
        dataset.add_update_resources(resourcesdata)
        with pytest.raises(NotRequestableError):
            dataset.get_fieldnames()
        with pytest.raises(NotRequestableError):
            dataset.add_fieldname("LALA")
        with pytest.raises(NotRequestableError):
            dataset.add_fieldnames(["LALA"])
        with pytest.raises(NotRequestableError):
            dataset.remove_fieldname("LALA")
        with pytest.raises(NotRequestableError):
            dataset.add_filetype("csv")
        with pytest.raises(NotRequestableError):
            dataset.add_filetypes(["csv"])
        with pytest.raises(NotRequestableError):
            dataset.remove_filetype("csv")
        assert dataset.get_filetypes() == ["xlsx", "csv", "xls"]

    def test_autocomplete(self, configuration, post_autocomplete):
        assert Dataset.autocomplete("acled") == dataset_autocomplete



"""
Calls a function that generates a dataset and creates it in HDX.

"""
import logging

from hdx.facades.simple import facade

from .my_code import generate_dataset

logger = logging.getLogger(__name__)


def main():
    """Generate dataset and create it in HDX"""

    dataset = generate_dataset()
    dataset.create_in_hdx()


if __name__ == "__main__":
    facade(main, hdx_site="stage")



from typing import Optional

from hdx.api.configuration import Configuration


class TestResult:
    actual_result = None


testresult = TestResult()


def my_testfn():
    testresult.actual_result = Configuration.read().get_hdx_site_url()


def my_testkeyfn():
    testresult.actual_result = Configuration.read().get_api_key()


def my_testuafn():
    testresult.actual_result = Configuration.read().user_agent


def my_excfn():
    testresult.actual_result = Configuration.read().get_hdx_site_url()
    raise ValueError("Some failure!")


def my_testfnkw(**kwargs):
    fn = kwargs.get("fn")
    if fn == "site":
        testresult.actual_result = Configuration.read().get_hdx_site_url()
    elif fn == "api":
        testresult.actual_result = Configuration.read().get_api_key()
    elif fn == "agent":
        testresult.actual_result = Configuration.read().user_agent
    elif fn == "exc":
        testresult.actual_result = Configuration.read().get_hdx_site_url()
        raise ValueError("Some failure!")


def my_testfnia(mydata: Optional[str] = None) -> str:
    testresult.actual_result = mydata



"""Facade to simplify project setup that calls project main function with kwargs"""
import logging
import sys
from inspect import BoundArguments
from typing import Any, Callable, List, Optional, Tuple

import defopt
from hdx.utilities.easy_logging import setup_logging
from hdx.utilities.useragent import UserAgent

from hdx.api import __version__
from hdx.api.configuration import Configuration

logger = logging.getLogger(__name__)
setup_logging(log_file="errors.log")


def _bind_known(fn: Callable) -> Tuple[Callable, BoundArguments, List[str]]:
    parser = defopt._create_parser(fn, cli_options="all")
    with defopt._colorama_text():
        args, argv = parser.parse_known_args()
        parsed_argv = vars(args)
    try:
        func = parsed_argv.pop("_func")
    except KeyError:
        # Workaround for http://bugs.python.org/issue9253#msg186387 (and
        # https://bugs.python.org/issue29298 which blocks using required=True).
        parser.error("too few arguments")
    sig = defopt.signature(func)
    ba = sig.bind_partial()
    ba.arguments.update(parsed_argv)
    return func, ba, argv


def _run_known(func: Callable, ba: BoundArguments) -> Any:
    sig = defopt.signature(func)
    (raises,) = [
        # typing_inspect does not allow fetching metadata; see e.g. ti#82.
        arg
        for arg in getattr(sig.return_annotation, "__metadata__", [])
        if isinstance(arg, defopt._Raises)
    ]
    # The function call should occur here to minimize effects on the traceback.
    try:
        return func(*ba.args, **ba.kwargs)
    except raises as e:
        sys.exit(e)


def _create_configuration(
    user_agent: Optional[str] = None,
    user_agent_config_yaml: Optional[str] = None,
    user_agent_lookup: Optional[str] = None,
    hdx_url: Optional[str] = None,
    hdx_site: Optional[str] = None,
    hdx_read_only: bool = False,
    hdx_key: Optional[str] = None,
    hdx_config_json: Optional[str] = None,
    hdx_config_yaml: Optional[str] = None,
    project_config_json: Optional[str] = None,
    project_config_yaml: Optional[str] = None,
    hdx_base_config_json: Optional[str] = None,
    hdx_base_config_yaml: Optional[str] = None,
) -> str:
    """
    Create HDX configuration

    Args:
        user_agent (Optional[str]): User agent string. HDXPythonLibrary/X.X.X- is prefixed. Must be supplied if remoteckan is not.
        user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
        user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.
        hdx_url (Optional[str]): HDX url to use. Overrides hdx_site.
        hdx_site (Optional[str]): HDX site to use eg. prod, test.
        hdx_read_only (bool): Whether to access HDX in read only mode. Defaults to False.
        hdx_key (Optional[str]): Your HDX key. Ignored if hdx_read_only = True.
        hdx_config_json (Optional[str]): Path to JSON HDX configuration OR
        hdx_config_yaml (Optional[str]): Path to YAML HDX configuration
        project_config_json (Optional[str]): Path to JSON Project configuration OR
        project_config_yaml (Optional[str]): Path to YAML Project configuration
        hdx_base_config_json (Optional[str]): Path to JSON HDX base configuration OR
        hdx_base_config_yaml (Optional[str]): Path to YAML HDX base configuration. Defaults to library's internal hdx_base_configuration.yml.

    Returns:
        str: HDX site url

    """
    arguments = locals()
    return Configuration._create(**arguments)


def facade(projectmainfn: Callable[[Any], None]):
    """Facade to simplify project setup that calls project main function. It infers
    command line arguments from the passed in function using defopt. The function passed
    in should have either type hints or a docstring from which to infer the command
    line arguments.

    Args:
        projectmainfn ((Any) -> None): main function of project

    Returns:
        None
    """

    #
    # Setting up configuration
    #
    func, ba, argv = _bind_known(projectmainfn)
    site_url = defopt.run(_create_configuration, argv=argv, cli_options="all")

    logger.info("--------------------------------------------------")
    logger.info(f"> Using HDX Python API Library {__version__}")
    logger.info(f"> HDX Site: {site_url}")

    UserAgent.user_agent = Configuration.read().user_agent
    _run_known(func, ba)



"""Facade to simplify project setup that calls project main function"""
import logging
from typing import Any, Callable

from hdx.utilities.easy_logging import setup_logging
from hdx.utilities.useragent import UserAgent

from hdx.api import __version__
from hdx.api.configuration import Configuration

logger = logging.getLogger(__name__)
setup_logging(log_file="errors.log")


def facade(projectmainfn: Callable[[None], None], **kwargs: Any):
    """Facade to simplify project setup that calls project main function

    Args:
        projectmainfn ((None) -> None): main function of project
        **kwargs: configuration parameters to pass to HDX Configuration class

    Returns:
        None
    """

    #
    # Setting up configuration
    #
    site_url = Configuration._create(**kwargs)

    logger.info("--------------------------------------------------")
    logger.info(f"> Using HDX Python API Library {__version__}")
    logger.info(f"> HDX Site: {site_url}")

    UserAgent.user_agent = Configuration.read().user_agent

    projectmainfn()



"""Configuration Tests"""
from os.path import join

import ckanapi
import pytest
from hdx.utilities.loader import LoadError
from hdx.utilities.useragent import UserAgentError

from hdx.api import __version__
from hdx.api.configuration import Configuration, ConfigurationError


class TestConfiguration:
    @pytest.fixture(scope="class")
    def hdx_base_config_yaml(self, configfolder):
        return join(configfolder, "hdx_base_config.yml")

    @pytest.fixture(scope="class")
    def hdx_base_config_json(self, configfolder):
        return join(configfolder, "hdx_base_config.json")

    @pytest.fixture(scope="class")
    def hdx_config_json(self, configfolder):
        return join(configfolder, "hdx_config.json")

    @pytest.fixture(scope="class")
    def hdx_missing_site_config_json(self, configfolder):
        return join(configfolder, "hdx_missing_site_config.json")

    @pytest.fixture(scope="class")
    def project_config_json(self, configfolder):
        return join(configfolder, "project_configuration.json")

    @pytest.fixture(scope="class")
    def user_agent_config_yaml(self, configfolder):
        return join(configfolder, "user_agent_config.yml")

    @pytest.fixture(scope="class")
    def user_agent_config2_yaml(self, configfolder):
        return join(configfolder, "user_agent_config2.yml")

    @pytest.fixture(scope="class")
    def user_agent_config3_yaml(self, configfolder):
        return join(configfolder, "user_agent_config3.yml")

    @pytest.fixture(scope="class")
    def empty_yaml(self, configfolder):
        return join(configfolder, "empty.yml")

    @pytest.fixture(scope="class")
    def user_agent_config_wrong_yaml(self, configfolder):
        return join(configfolder, "user_agent_config_wrong.yml")

    def test_init(
        self,
        hdx_config_json,
        hdx_config_yaml,
        hdx_base_config_yaml,
        hdx_base_config_json,
        hdx_missing_site_config_json,
        project_config_json,
        project_config_yaml,
    ):
        default_config_file = Configuration.default_hdx_config_yaml
        Configuration.default_hdx_config_yaml = "NOT EXIST"
        with pytest.raises(UserAgentError):
            Configuration()
        Configuration.default_hdx_config_yaml = default_config_file

        Configuration.default_hdx_config_yaml = hdx_config_yaml
        assert Configuration(user_agent="test").get_api_key() == "12345"
        Configuration.default_hdx_config_yaml = default_config_file

        with pytest.raises(IOError):
            Configuration(
                hdx_config_yaml="NOT_EXIST",
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(IOError):
            Configuration(
                hdx_config_yaml=hdx_config_yaml,
                hdx_base_config_yaml="NOT_EXIST",
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(IOError):
            Configuration(
                hdx_config_yaml=hdx_config_yaml,
                hdx_base_config_json="NOT_EXIST",
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                hdx_config_dict={"a": 1},
                hdx_config_yaml=hdx_base_config_yaml,
                hdx_base_config_json=hdx_base_config_json,
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                hdx_config_dict={"a": 1},
                hdx_config_json=hdx_config_json,
                hdx_base_config_json=hdx_base_config_json,
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                hdx_config_json=hdx_config_json,
                hdx_config_yaml=hdx_config_yaml,
                hdx_base_config_json=hdx_base_config_json,
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                hdx_config_yaml=hdx_config_yaml,
                hdx_base_config_dict={"a": 1},
                hdx_base_config_yaml=hdx_base_config_yaml,
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                hdx_config_yaml=hdx_config_yaml,
                hdx_base_config_dict={"a": 1},
                hdx_base_config_json=hdx_base_config_json,
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                hdx_config_yaml=hdx_config_yaml,
                hdx_base_config_json=hdx_base_config_json,
                hdx_base_config_yaml=hdx_base_config_yaml,
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(IOError):
            Configuration(
                hdx_config_yaml=hdx_config_yaml,
                project_config_yaml="NOT_EXIST",
            )

        with pytest.raises(IOError):
            Configuration(
                hdx_config_yaml=hdx_config_yaml,
                project_config_json="NOT_EXIST",
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                user_agent="test",
                hdx_config_json=hdx_missing_site_config_json,
                project_config_json=project_config_json,
                hdx_base_config_json=hdx_base_config_json,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                user_agent="test",
                hdx_site="NOT_EXIST",
                hdx_config_yaml=hdx_config_yaml,
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                user_agent="test",
                hdx_config_yaml=hdx_config_yaml,
                project_config_json=project_config_json,
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                user_agent="test",
                hdx_config_yaml=hdx_config_yaml,
                project_config_dict={"la": "la"},
                project_config_yaml=project_config_yaml,
            )

        with pytest.raises(ConfigurationError):
            Configuration(
                user_agent="test",
                hdx_config_yaml=hdx_config_yaml,
                project_config_dict={"la": "la"},
                project_config_json=project_config_json,
            )

    def test_hdx_configuration_dict(self, project_config_yaml, mocksmtp):
        Configuration._create(
            user_agent="test",
            hdx_config_dict={
                "hdx_site": "prod",
                "hdx_read_only": True,
                "hdx_key": "abcde",
            },
            hdx_base_config_dict={
                "hdx_prod_site": {
                    "url": "https://data.humdata.org",
                },
                "XYZ": {"567": 987},
            },
            project_config_yaml=project_config_yaml,
        )
        expected_configuration = {
            "hdx_site": "prod",
            "hdx_read_only": True,
            "hdx_key": "abcde",
            "hdx_prod_site": {
                "url": "https://data.humdata.org",
            },
            "XYZ": {"567": 987},
            "approved_tags_vocabulary": "Topics",
            "tags_list_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/Accepted_Tags.csv",
            "tags_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/Tag_Mapping.csv",
            "formats_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/resource_formats.json",
        }

        configuration = Configuration.read()
        assert configuration == expected_configuration
        assert (
            configuration.get_user_agent()
            == f"HDXPythonLibrary/{__version__}-test"
        )

        Configuration._create(
            hdx_config_dict={
                "hdx_site": "prod",
                "hdx_read_only": True,
                "hdx_key": "abcde",
                "user_agent": "test",
            },
            hdx_base_config_dict={
                "hdx_prod_site": {
                    "url": "https://data.humdata.org",
                },
                "XYZ": {"567": 987},
            },
            project_config_yaml=project_config_yaml,
        )
        configuration = Configuration.read()
        expected_configuration["user_agent"] = "test"
        assert configuration == expected_configuration
        assert (
            configuration.get_user_agent()
            == f"HDXPythonLibrary/{__version__}-test"
        )

        smtp_initargs = {
            "host": "localhost",
            "port": 123,
            "local_hostname": "mycomputer.fqdn.com",
            "timeout": 3,
            "source_address": ("machine", 456),
        }
        username = "user"
        password = "pass"
        email_config_dict = {
            "connection_type": "ssl",
            "username": username,
            "password": password,
        }
        email_config_dict.update(smtp_initargs)

        recipients = ["larry@gmail.com", "moe@gmail.com", "curly@gmail.com"]
        subject = "hello"
        body = "hello there"
        sender = "me@gmail.com"
        mail_options = ["a", "b"]
        rcpt_options = [1, 2]

        with pytest.raises(ConfigurationError):
            configuration.emailer()
        configuration.setup_emailer(email_config_dict=email_config_dict)
        email = configuration.emailer()
        email.send(
            recipients,
            subject,
            body,
            sender=sender,
            mail_options=mail_options,
            rcpt_options=rcpt_options,
        )
        assert email.server.type == "smtpssl"
        assert email.server.initargs == smtp_initargs
        assert email.server.username == username
        assert email.server.password == password
        assert email.server.sender == sender
        assert email.server.recipients == recipients
        assert (
            email.server.msg
            == """Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Subject: hello
From: me@gmail.com
To: larry@gmail.com, moe@gmail.com, curly@gmail.com

hello there"""
        )
        assert email.server.send_args == {
            "mail_options": ["a", "b"],
            "rcpt_options": [1, 2],
        }

    def test_hdx_configuration_json(
        self, hdx_config_json, hdx_base_config_json, project_config_yaml
    ):
        Configuration._create(
            user_agent="test",
            hdx_config_json=hdx_config_json,
            hdx_base_config_json=hdx_base_config_json,
            project_config_yaml=project_config_yaml,
        )
        expected_configuration = {
            "hdx_site": "stage",
            "hdx_read_only": False,
            "hdx_key_stage": "54321",
            "hdx_prod_site": {
                "url": "https://data.humdata.org",
            },
            "hdx_stage_site": {
                "url": "https://stage.data-humdata-org.ahconu.org",
                "username": "tumteetum",
                "password": "tumteetumteetum",
            },
            "dataset": {
                "required_fields": [
                    "name",
                    "dataset_date",
                ]
            },
            "resource": {"required_fields": ["name", "description"]},
            "showcase": {"required_fields": ["name"]},
            "approved_tags_vocabulary": "Topics",
            "tags_list_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/Accepted_Tags.csv",
            "tags_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/Tag_Mapping.csv",
            "formats_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/resource_formats.json",
        }
        configuration = Configuration.read()
        assert configuration == expected_configuration
        assert configuration.hdx_site == "hdx_stage_site"
        assert configuration.hdx_key == "54321"

    def test_hdx_configuration_yaml(
        self, hdx_config_yaml, hdx_base_config_yaml, project_config_yaml
    ):
        Configuration._create(
            user_agent="test",
            hdx_config_yaml=hdx_config_yaml,
            hdx_base_config_yaml=hdx_base_config_yaml,
            project_config_yaml=project_config_yaml,
        )
        expected_configuration = {
            "hdx_site": "prod",
            "hdx_read_only": False,
            "hdx_key": "12345",
            "hdx_prod_site": {
                "url": "https://data.humdata.org",
            },
            "hdx_stage_site": {
                "url": "https://stage.data-humdata-org.ahconu.org",
                "username": "lala",
                "password": "lalala",
            },
            "dataset": {
                "required_fields": [
                    "name",
                    "title",
                    "dataset_date",
                ]
            },
            "resource": {
                "required_fields": ["package_id", "name", "description"]
            },
            "showcase": {"required_fields": ["name", "title"]},
            "approved_tags_vocabulary": "Topics",
            "tags_list_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/Accepted_Tags.csv",
            "tags_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/Tag_Mapping.csv",
            "formats_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/resource_formats.json",
        }
        configuration = Configuration.read()
        assert configuration == expected_configuration
        assert configuration.hdx_site == "hdx_prod_site"
        assert configuration.hdx_key == "12345"

    def test_project_configuration_dict(self, hdx_config_yaml):
        Configuration._create(
            user_agent="test", hdx_config_yaml=hdx_config_yaml
        )
        expected_configuration = {
            "hdx_site": "prod",
            "hdx_read_only": False,
            "hdx_key": "12345",
            "hdx_prod_site": {
                "url": "https://data.humdata.org",
            },
            "hdx_demo_site": {
                "url": "https://demo.data-humdata-org.ahconu.org",
            },
            "hdx_stage_site": {
                "url": "https://stage.data-humdata-org.ahconu.org",
            },
            "hdx_feature_site": {
                "url": "https://feature.data-humdata-org.ahconu.org",
            },
            "hdx_dev_site": {
                "url": "https://dev.data-humdata-org.ahconu.org",
            },
            "dataset": {
                "required_fields": [
                    "name",
                    "private",
                    "title",
                    "notes",
                    "dataset_source",
                    "owner_org",
                    "maintainer",
                    "dataset_date",
                    "data_update_frequency",
                    "groups",
                    "license_id",
                    "methodology",
                    "tags",
                ]
            },
            "dataset-requestable": {
                "required_fields": [
                    "name",
                    "title",
                    "notes",
                    "dataset_source",
                    "owner_org",
                    "maintainer",
                    "dataset_date",
                    "data_update_frequency",
                    "groups",
                    "tags",
                    "field_names",
                    "file_types",
                    "num_of_rows",
                ]
            },
            "resource": {
                "required_fields": [
                    "package_id",
                    "name",
                    "format",
                    "description",
                    "url_type",
                    "resource_type",
                ]
            },
            "showcase": {
                "required_fields": [
                    "name",
                    "title",
                    "notes",
                    "url",
                    "image_url",
                    "tags",
                ]
            },
            "user": {
                "required_fields": [
                    "name",
                    "email",
                    "password",
                    "fullname",
                    "about",
                ],
                "ignore_on_update": "password",
            },
            "organization": {
                "required_fields": [
                    "name",
                    "title",
                    "description",
                ]
            },
            "resource view": {
                "required_fields": [
                    "resource_id",
                    "title",
                    "view_type",
                ]
            },
            "vocabulary": {
                "required_fields": [
                    "name",
                    "tags",
                ]
            },
            "approved_tags_vocabulary": "Topics",
            "tags_list_url": "https://docs.google.com/spreadsheets/d/e/2PACX-1vRjeajloIuQl8mfTSHU71ZgbHSgYYUgHrLqyjHSuQJ-zMqS3SVM9hJqMs72L-84LQ/pub?gid=1739051517&single=true&output=csv",
            "tags_mapping_url": "https://docs.google.com/spreadsheets/d/e/2PACX-1vRjeajloIuQl8mfTSHU71ZgbHSgYYUgHrLqyjHSuQJ-zMqS3SVM9hJqMs72L-84LQ/pub?gid=334970416&single=true&output=csv",
            "formats_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-ckan/master/ckanext-hdx_package/ckanext/hdx_package/config/resource_formats.json",
        }
        assert Configuration.read() == expected_configuration
        Configuration._create(
            user_agent="test",
            hdx_config_yaml=hdx_config_yaml,
            project_config_dict={"abc": "123"},
        )
        expected_configuration["abc"] = "123"
        assert Configuration.read() == expected_configuration

    def test_project_configuration_json(
        self, hdx_config_yaml, project_config_json
    ):
        Configuration._create(
            user_agent="test",
            hdx_config_yaml=hdx_config_yaml,
            project_config_json=project_config_json,
        )
        expected_configuration = {
            "hdx_site": "prod",
            "hdx_read_only": False,
            "hdx_key": "12345",
            "hdx_prod_site": {
                "url": "https://data.humdata.org",
            },
            "hdx_demo_site": {
                "url": "https://demo.data-humdata-org.ahconu.org",
            },
            "hdx_stage_site": {
                "url": "https://stage.data-humdata-org.ahconu.org",
            },
            "hdx_feature_site": {
                "url": "https://feature.data-humdata-org.ahconu.org",
            },
            "hdx_dev_site": {
                "url": "https://dev.data-humdata-org.ahconu.org",
            },
            "my_param": "abc",
            "dataset": {
                "required_fields": [
                    "name",
                    "private",
                    "title",
                    "notes",
                    "dataset_source",
                    "owner_org",
                    "maintainer",
                    "dataset_date",
                    "data_update_frequency",
                    "groups",
                    "license_id",
                    "methodology",
                    "tags",
                ]
            },
            "dataset-requestable": {
                "required_fields": [
                    "name",
                    "title",
                    "notes",
                    "dataset_source",
                    "owner_org",
                    "maintainer",
                    "dataset_date",
                    "data_update_frequency",
                    "groups",
                    "tags",
                    "field_names",
                    "file_types",
                    "num_of_rows",
                ]
            },
            "resource": {
                "required_fields": [
                    "package_id",
                    "name",
                    "format",
                    "description",
                    "url_type",
                    "resource_type",
                ]
            },
            "showcase": {
                "required_fields": [
                    "name",
                    "title",
                    "notes",
                    "url",
                    "image_url",
                    "tags",
                ]
            },
            "user": {
                "required_fields": [
                    "name",
                    "email",
                    "password",
                    "fullname",
                    "about",
                ],
                "ignore_on_update": "password",
            },
            "organization": {
                "required_fields": [
                    "name",
                    "title",
                    "description",
                ]
            },
            "resource view": {
                "required_fields": [
                    "resource_id",
                    "title",
                    "view_type",
                ]
            },
            "vocabulary": {
                "required_fields": [
                    "name",
                    "tags",
                ]
            },
            "approved_tags_vocabulary": "Topics",
            "tags_list_url": "https://docs.google.com/spreadsheets/d/e/2PACX-1vRjeajloIuQl8mfTSHU71ZgbHSgYYUgHrLqyjHSuQJ-zMqS3SVM9hJqMs72L-84LQ/pub?gid=1739051517&single=true&output=csv",
            "tags_mapping_url": "https://docs.google.com/spreadsheets/d/e/2PACX-1vRjeajloIuQl8mfTSHU71ZgbHSgYYUgHrLqyjHSuQJ-zMqS3SVM9hJqMs72L-84LQ/pub?gid=334970416&single=true&output=csv",
            "formats_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-ckan/master/ckanext-hdx_package/ckanext/hdx_package/config/resource_formats.json",
        }
        assert Configuration.read() == expected_configuration

    def test_project_configuration_yaml(
        self, hdx_config_yaml, project_config_yaml
    ):
        Configuration._create(
            user_agent="test",
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
        )
        expected_configuration = {
            "hdx_site": "prod",
            "hdx_read_only": False,
            "hdx_key": "12345",
            "hdx_prod_site": {
                "url": "https://data.humdata.org",
            },
            "hdx_demo_site": {
                "url": "https://demo.data-humdata-org.ahconu.org",
            },
            "hdx_stage_site": {
                "url": "https://stage.data-humdata-org.ahconu.org",
            },
            "hdx_feature_site": {
                "url": "https://feature.data-humdata-org.ahconu.org",
            },
            "hdx_dev_site": {
                "url": "https://dev.data-humdata-org.ahconu.org",
            },
            "dataset": {
                "required_fields": [
                    "name",
                    "private",
                    "title",
                    "notes",
                    "dataset_source",
                    "owner_org",
                    "maintainer",
                    "dataset_date",
                    "data_update_frequency",
                    "groups",
                    "license_id",
                    "methodology",
                    "tags",
                ]
            },
            "dataset-requestable": {
                "required_fields": [
                    "name",
                    "title",
                    "notes",
                    "dataset_source",
                    "owner_org",
                    "maintainer",
                    "dataset_date",
                    "data_update_frequency",
                    "groups",
                    "tags",
                    "field_names",
                    "file_types",
                    "num_of_rows",
                ]
            },
            "resource": {
                "required_fields": [
                    "package_id",
                    "name",
                    "format",
                    "description",
                    "url_type",
                    "resource_type",
                ]
            },
            "showcase": {
                "required_fields": [
                    "name",
                    "title",
                    "notes",
                    "url",
                    "image_url",
                    "tags",
                ]
            },
            "user": {
                "required_fields": [
                    "name",
                    "email",
                    "password",
                    "fullname",
                    "about",
                ],
                "ignore_on_update": "password",
            },
            "organization": {
                "required_fields": [
                    "name",
                    "title",
                    "description",
                ]
            },
            "resource view": {
                "required_fields": [
                    "resource_id",
                    "title",
                    "view_type",
                ]
            },
            "vocabulary": {
                "required_fields": [
                    "name",
                    "tags",
                ]
            },
            "approved_tags_vocabulary": "Topics",
            "tags_list_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/Accepted_Tags.csv",
            "tags_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/Tag_Mapping.csv",
            "formats_mapping_url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/resource_formats.json",
        }
        assert Configuration.read() == expected_configuration

    def test_get_hdx_key_site(self, hdx_config_yaml, project_config_yaml):
        Configuration._create(
            user_agent="test",
            hdx_config_yaml=hdx_config_yaml,
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        actual_configuration = Configuration.read()
        assert actual_configuration.get_api_key() == "12345"
        assert (
            actual_configuration.get_hdx_site_url()
            == "https://data.humdata.org"
        )
        assert actual_configuration._get_credentials() is None
        assert (
            actual_configuration.get_dataset_url("mydataset")
            == "https://data.humdata.org/dataset/mydataset"
        )

    def test_set_hdx_key_value(self, project_config_yaml):
        Configuration._create(
            user_agent="test",
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        configuration = Configuration.read()
        assert configuration.get_api_key() == "TEST_HDX_KEY"
        configuration.set_api_key("NEW API KEY")
        assert configuration.get_api_key() == "NEW API KEY"
        Configuration._create(
            user_agent="test",
            hdx_site="prod",
            hdx_read_only=True,
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        assert Configuration.read().get_api_key() is None
        configuration = Configuration.read()
        configuration.set_api_key("TEST API KEY")
        assert configuration.get_api_key() is None
        configuration.set_read_only(False)
        assert configuration.get_api_key() == "TEST API KEY"
        configuration.set_read_only(True)
        assert configuration.get_api_key() is None
        configuration.set_api_key("NEW API KEY")
        configuration.set_read_only(False)
        assert configuration.get_api_key() == "NEW API KEY"

    def test_create_set_configuration(self, project_config_yaml):
        Configuration._create(
            user_agent="test",
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        with pytest.raises(ConfigurationError):
            Configuration.create(
                user_agent="test",
                hdx_site="prod",
                hdx_key="TEST_HDX_KEY",
                hdx_base_config_dict={},
                project_config_yaml=project_config_yaml,
            )
        configuration = Configuration(
            user_agent="test",
            hdx_site="stage",
            hdx_key="OTHER_TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        Configuration.setup(configuration)
        assert Configuration.read() == configuration
        Configuration.delete()
        with pytest.raises(ConfigurationError):
            Configuration.read()
        Configuration.create(
            user_agent="test",
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        assert Configuration.read().get_api_key() == "TEST_HDX_KEY"

    def test_remoteckan_validlocations(self, project_config_yaml):
        Configuration._create(
            user_agent="test",
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        remoteckan = ckanapi.RemoteCKAN(
            "http://lalala", apikey="12345", user_agent="HDXPythonLibrary/1.0"
        )
        Configuration.read().setup_session_remoteckan(remoteckan=remoteckan)
        assert Configuration.read().remoteckan() == remoteckan
        remoteckan = ckanapi.RemoteCKAN(
            "http://hahaha", apikey="54321", user_agent="HDXPythonLibrary/0.5"
        )
        Configuration._create(
            user_agent="test",
            remoteckan=remoteckan,
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        assert Configuration.read().remoteckan() == remoteckan
        Configuration.read()._remoteckan = None
        with pytest.raises(ConfigurationError):
            Configuration.read().remoteckan()
        Configuration.delete()
        with pytest.raises(ConfigurationError):
            Configuration.read().remoteckan()

    def test_user_agent(
        self,
        user_agent_config_yaml,
        user_agent_config2_yaml,
        user_agent_config3_yaml,
        empty_yaml,
        user_agent_config_wrong_yaml,
        project_config_yaml,
    ):
        Configuration._create(
            user_agent_config_yaml=user_agent_config_yaml,
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        assert (
            Configuration.read().remoteckan().user_agent
            == f"lala:HDXPythonLibrary/{__version__}-myua"
        )
        Configuration._create(
            user_agent_config_yaml=user_agent_config2_yaml,
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        assert (
            Configuration.read().remoteckan().user_agent
            == f"HDXPythonLibrary/{__version__}-myuseragent"
        )
        Configuration._create(
            user_agent_config_yaml=user_agent_config3_yaml,
            user_agent_lookup="lookup",
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        assert (
            Configuration.read().remoteckan().user_agent
            == f"HDXPythonLibrary/{__version__}-mylookupagent"
        )
        Configuration._create(
            user_agent_config_yaml=user_agent_config3_yaml,
            user_agent_lookup="lookup2",
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        assert (
            Configuration.read().remoteckan().user_agent
            == f"HDXPythonLibrary/{__version__}-mylookupagent2"
        )
        Configuration._create(
            user_agent="my_ua",
            preprefix="papa",
            hdx_site="prod",
            hdx_key="TEST_HDX_KEY",
            hdx_base_config_dict={},
            project_config_yaml=project_config_yaml,
        )
        assert (
            Configuration.read().remoteckan().user_agent
            == f"papa:HDXPythonLibrary/{__version__}-my_ua"
        )
        with pytest.raises(UserAgentError):
            Configuration._create(
                user_agent_config_yaml=user_agent_config3_yaml,
                user_agent_lookup="fail",
                hdx_site="prod",
                hdx_key="TEST_HDX_KEY",
                hdx_base_config_dict={},
                project_config_yaml=project_config_yaml,
            )
        with pytest.raises(LoadError):
            Configuration._create(
                user_agent_config_yaml=empty_yaml,
                hdx_site="prod",
                hdx_key="TEST_HDX_KEY",
                hdx_base_config_dict={},
                project_config_yaml=project_config_yaml,
            )
        with pytest.raises(UserAgentError):
            Configuration._create(
                user_agent_config_yaml=user_agent_config_wrong_yaml,
                hdx_site="prod",
                hdx_key="TEST_HDX_KEY",
                hdx_base_config_dict={},
                project_config_yaml=project_config_yaml,
            )
        with pytest.raises(UserAgentError):
            Configuration._create(
                hdx_site="prod",
                hdx_key="TEST_HDX_KEY",
                hdx_base_config_dict={},
                project_config_yaml=project_config_yaml,
            )



"""HDXObject abstract class containing helper functions for creating, checking, and updating HDX objects.
New HDX objects should extend this in similar fashion to Resource for example.
"""
import copy
import logging
from abc import ABC, abstractmethod
from collections import UserDict
from typing import Any, Dict, List, Optional, Tuple, Union

from ckanapi.errors import NotFound
from hdx.utilities.dictandlist import merge_two_dictionaries
from hdx.utilities.loader import (
    load_json_into_existing_dict,
    load_yaml_into_existing_dict,
)

from hdx.api.configuration import Configuration

logger = logging.getLogger(__name__)


class HDXError(Exception):
    pass


class HDXObject(UserDict, ABC):
    """HDXObject abstract class containing helper functions for creating, checking, and updating HDX objects.
    New HDX objects should extend this in similar fashion to Resource for example.

    Args:
        initial_data (Dict): Initial metadata dictionary
        configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.
    """

    @staticmethod
    @abstractmethod
    def actions() -> Dict[str, str]:
        """Dictionary of actions that can be performed on object

        Returns:
            Dict[str, str]: Dictionary of actions that can be performed on object
        """

    def __init__(
        self, initial_data: Dict, configuration: Optional[Configuration] = None
    ) -> None:
        self.old_data = None
        if configuration is None:
            self.configuration: Configuration = Configuration.read()
        else:
            self.configuration: Configuration = configuration
        super().__init__(initial_data)

    def get_old_data_dict(self) -> None:
        """Get previous internal dictionary

        Returns:
            dict: Previous internal dictionary
        """
        return self.old_data

    def update_from_yaml(self, path: str) -> None:
        """Update metadata with static metadata from YAML file

        Args:
            path (str): Path to YAML dataset metadata

        Returns:
            None
        """
        self.data = load_yaml_into_existing_dict(self.data, path)

    def update_from_json(self, path: str) -> None:
        """Update metadata with static metadata from JSON file

        Args:
            path (str): Path to JSON dataset metadata

        Returns:
            None
        """
        self.data = load_json_into_existing_dict(self.data, path)

    def _read_from_hdx(
        self,
        object_type: str,
        value: str,
        fieldname: str = "id",
        action: Optional[str] = None,
        **kwargs: Any,
    ) -> Tuple[bool, Union[Dict, str]]:
        """Makes a read call to HDX passing in given parameter.

        Args:
            object_type (str): Description of HDX object type (for messages)
            value (str): Value of HDX field
            fieldname (str): HDX field name. Defaults to id.
            action (Optional[str]): Replacement CKAN action url to use. Defaults to None.
            **kwargs: Other fields to pass to CKAN.

        Returns:
            Tuple[bool, Union[Dict, str]]: (True/False, HDX object metadata/Error)
        """
        if not fieldname:
            raise HDXError(f"Empty {object_type} field name!")
        if action is None:
            action = self.actions()["show"]
        data = {fieldname: value}
        data.update(kwargs)
        try:
            result = self.configuration.call_remoteckan(action, data)
            return True, result
        except NotFound:
            return False, f"{fieldname}={value}: not found!"
        except Exception as e:
            raise HDXError(
                f"Failed when trying to read: {fieldname}={value}! (POST)"
            ) from e

    def _load_from_hdx(self, object_type: str, id_field: str) -> bool:
        """Helper method to load the HDX object given by identifier from HDX

        Args:
            object_type (str): Description of HDX object type (for messages)
            id_field (str): HDX object identifier

        Returns:
            bool: True if loaded, False if not
        """
        success, result = self._read_from_hdx(object_type, id_field)
        if success:
            self.old_data = self.data
            self.data = result
            return True
        logger.debug(result)
        return False

    @staticmethod
    @abstractmethod
    def read_from_hdx(
        id_field: str, configuration: Optional[Configuration] = None
    ) -> Optional["HDXObject"]:
        """Abstract method to read the HDX object given by identifier from HDX and return it

        Args:
            id_field (str): HDX object identifier
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Optional[HDXObject]: HDX object if successful read, None if not
        """

    @classmethod
    def _read_from_hdx_class(
        cls,
        object_type: str,
        identifier: str,
        configuration: Optional[Configuration] = None,
    ) -> Optional["HDXObject"]:
        """Reads the HDX object given by identifier from HDX and returns it

        Args:
            object_type (str): Description of HDX object type (for messages)
            identifier (str): Identifier
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Optional[HDXObject]: HDX object if successful read, None if not
        """
        hdxobject = cls(configuration=configuration)
        result = hdxobject._load_from_hdx(object_type, identifier)
        if result:
            return hdxobject
        return None

    def _check_existing_object(
        self, object_type: str, id_field_name: str
    ) -> None:
        if not self.data:
            raise HDXError(f"No data in {object_type}!")
        if id_field_name not in self.data:
            raise HDXError(
                f"No {id_field_name} field (mandatory) in {object_type}!"
            )

    def _check_load_existing_object(
        self, object_type: str, id_field_name: str, operation: str = "update"
    ) -> None:
        """Check metadata exists and contains HDX object identifier, and if so load HDX object

        Args:
            object_type (str): Description of HDX object type (for messages)
            id_field_name (str): Name of field containing HDX object identifier
            operation (str): Operation to report if error. Defaults to update.

        Returns:
            None
        """
        self._check_existing_object(object_type, id_field_name)
        if not self._load_from_hdx(object_type, self.data[id_field_name]):
            raise HDXError(f"No existing {object_type} to {operation}!")

    @abstractmethod
    def check_required_fields(self, ignore_fields: List[str] = list()) -> None:
        """Abstract method to check that metadata for HDX object is complete. The parameter ignore_fields should
        be set if required to any fields that should be ignored for the particular operation.

        Args:
            ignore_fields (List[str]): Fields to ignore. Default is [].

        Returns:
            None
        """

    def _check_required_fields(
        self, object_type: str, ignore_fields: List[str]
    ) -> None:
        """Helper method to check that metadata for HDX object is complete

        Args:
            ignore_fields (List[str]): Any fields to ignore in the check

        Returns:
            None
        """
        for field in self.configuration[object_type]["required_fields"]:
            if field not in ignore_fields:
                if field not in self.data:
                    raise HDXError(
                        f"Field {field} is missing in {object_type}!"
                    )
                if not self.data[field] and not isinstance(
                    self.data[field], bool
                ):
                    raise HDXError(f"Field {field} is empty in {object_type}!")

    def _check_kwargs_fields(self, object_type: str, **kwargs: Any) -> None:
        """Helper method to check kwargs and set fields appropriately and to check metadata fields unless it is
        specified not to do so.

        Args:
            object_type (str): Description of HDX object type (for messages)
            **kwargs: See below
            ignore_field (str): Any field to ignore when checking dataset metadata. Defaults to None.

        Returns:
            None
        """
        if (
            "batch_mode" in kwargs
        ):  # Whether or not CKAN should change groupings of datasets on /datasets page
            self.data["batch_mode"] = kwargs["batch_mode"]
        if (
            "skip_validation" in kwargs
        ):  # Whether or not CKAN should perform validation steps (checking fields present)
            self.data["skip_validation"] = kwargs["skip_validation"]
        if "ignore_check" not in kwargs or not kwargs.get(
            "ignore_check"
        ):  # allow ignoring of field checks
            ignore_fields = kwargs.get("ignore_fields", list())
            ignore_field = self.configuration[object_type].get(
                "ignore_on_update"
            )
            if ignore_field and ignore_field not in ignore_fields:
                ignore_fields.append(ignore_field)
            ignore_field = kwargs.get("ignore_field")
            if ignore_field and ignore_field not in ignore_fields:
                ignore_fields.append(ignore_field)
            self.check_required_fields(ignore_fields=ignore_fields)

    def _hdx_update(
        self,
        object_type: str,
        id_field_name: str,
        files_to_upload: Dict = dict(),
        force_active: bool = False,
        **kwargs: Any,
    ) -> None:
        """Helper method to update HDX object

        Args:
            object_type (str): Description of HDX object type (for messages)
            id_field_name (str): Name of field containing HDX object identifier
            files_to_upload (Dict): Files to upload to HDX
            force_active (bool): Make object state active. Defaults to False.
            **kwargs: See below
            operation (str): Operation to perform eg. patch. Defaults to update.
            ignore_field (str): Any field to ignore when checking dataset metadata. Defaults to None.

        Returns:
            None
        """
        self._check_kwargs_fields(object_type, **kwargs)
        operation = kwargs.get("operation", "update")
        self._save_to_hdx(
            operation, id_field_name, files_to_upload, force_active
        )

    def _merge_hdx_update(
        self,
        object_type: str,
        id_field_name: str,
        files_to_upload: Dict = dict(),
        force_active: bool = False,
        **kwargs: Any,
    ) -> None:
        """Helper method to check if HDX object exists and update it

        Args:
            object_type (str): Description of HDX object type (for messages)
            id_field_name (str): Name of field containing HDX object identifier
            files_to_upload (Dict): Files to upload to HDX
            force_active (bool): Make object state active. Defaults to False.
            **kwargs: See below
            operation (str): Operation to perform eg. patch. Defaults to update.
            ignore_field (str): Any field to ignore when checking dataset metadata. Defaults to None.

        Returns:
            None
        """
        merge_two_dictionaries(self.data, self.old_data)
        self._hdx_update(
            object_type,
            id_field_name,
            files_to_upload=files_to_upload,
            force_active=force_active,
            **kwargs,
        )

    @abstractmethod
    def update_in_hdx(self) -> None:
        """Abstract method to check if HDX object exists in HDX and if so, update it

        Returns:
            None
        """

    def _update_in_hdx(
        self,
        object_type: str,
        id_field_name: str,
        files_to_upload: Dict = dict(),
        force_active: bool = True,
        **kwargs: Any,
    ) -> None:
        """Helper method to check if HDX object exists in HDX and if so, update it

        Args:
            object_type (str): Description of HDX object type (for messages)
            id_field_name (str): Name of field containing HDX object identifier
            files_to_upload (Dict): Files to upload to HDX
            force_active (bool): Make object state active. Defaults to True.
            **kwargs: See below
            operation (str): Operation to perform eg. patch. Defaults to update.
            ignore_field (str): Any field to ignore when checking dataset metadata. Defaults to None.

        Returns:
            None
        """

        self._check_load_existing_object(object_type, id_field_name)
        # We load an existing object even though it may well have been loaded already
        # to prevent an admittedly unlikely race condition where someone has updated
        # the object in the intervening time
        self._merge_hdx_update(
            object_type,
            id_field_name,
            files_to_upload,
            force_active=force_active,
            **kwargs,
        )

    def _write_to_hdx(
        self,
        action: str,
        data: Dict,
        id_field_name: str = None,
        files_to_upload: Dict = dict(),
    ) -> Union[Dict, List]:
        """Creates or updates an HDX object in HDX and return HDX object metadata dict

        Args:
            action (str): Action to perform eg. 'create', 'update'
            data (Dict): Data to write to HDX
            id_field_name (Optional[str]): Name of field containing HDX object identifier. Defaults to None.
            files_to_upload (Dict): Files to upload to HDX

        Returns:
            Union[Dict,List]: HDX object metadata
        """
        try:
            for key, value in files_to_upload.items():
                files_to_upload[key] = open(value, "rb")
            return self.configuration.call_remoteckan(
                self.actions()[action], data, files=files_to_upload
            )
        except Exception as e:
            if id_field_name:
                idstr = f" {data[id_field_name]}"
            else:
                idstr = ""
            raise HDXError(
                f"Failed when trying to {action}{idstr}! (POST)"
            ) from e
        finally:
            for file in files_to_upload.values():
                if isinstance(file, str):
                    continue
                file.close()

    def _save_to_hdx(
        self,
        action: str,
        id_field_name: str,
        files_to_upload: Dict = dict(),
        force_active: bool = False,
    ) -> None:
        """Creates or updates an HDX object in HDX, saving current data and replacing with returned HDX object data
        from HDX

        Args:
            action (str): Action to perform: 'create' or 'update'
            id_field_name (str): Name of field containing HDX object identifier
            files_to_upload (Dict): Files to upload to HDX
            force_active (bool): Make object state active. Defaults to False.

        Returns:
            None
        """
        if force_active:
            self.data["state"] = "active"
        result = self._write_to_hdx(
            action, self.data, id_field_name, files_to_upload
        )
        self.old_data = self.data
        self.data = result

    @abstractmethod
    def create_in_hdx(self) -> None:
        """Abstract method to check if resource exists in HDX and if so, update it, otherwise create it

        Returns:
            None
        """

    def _create_in_hdx(
        self,
        object_type: str,
        id_field_name: str,
        name_field_name: str,
        files_to_upload: Dict = dict(),
        force_active: bool = True,
        **kwargs: Any,
    ) -> None:
        """Helper method to check if resource exists in HDX and if so, update it, otherwise create it


        Args:
            object_type (str): Description of HDX object type (for messages)
            id_field_name (str): Name of field containing HDX object identifier
            name_field_name (str): Name of field containing HDX object name
            files_to_upload (Dict): Files to upload to HDX
            force_active (bool): Make object state active. Defaults to True.

        Returns:
            None
        """
        if "ignore_check" not in kwargs:  # allow ignoring of field checks
            self.check_required_fields()
        if id_field_name in self.data and self._load_from_hdx(
            object_type, self.data[id_field_name]
        ):
            logger.warning(
                f"{object_type} exists. Updating {self.data[id_field_name]}"
            )
            self._merge_hdx_update(
                object_type,
                id_field_name,
                files_to_upload,
                force_active,
                **kwargs,
            )
        else:
            self._save_to_hdx(
                "create", name_field_name, files_to_upload, force_active
            )

    @abstractmethod
    def delete_from_hdx(self) -> None:
        """Abstract method to deletes a resource from HDX

        Returns:
            None
        """

    def _delete_from_hdx(self, object_type: str, id_field_name: str) -> None:
        """Helper method to deletes a resource from HDX

        Args:
            object_type (str): Description of HDX object type (for messages)
            id_field_name (str): Name of field containing HDX object identifier

        Returns:
            None
        """
        if id_field_name not in self.data:
            raise HDXError(
                f"No {id_field_name} field (mandatory) in {object_type}!"
            )
        self._save_to_hdx("delete", id_field_name)

    @classmethod
    def _autocomplete(
        cls,
        name: str,
        limit: int = 20,
        configuration: Optional[Configuration] = None,
        **kwargs: Any,
    ) -> List:
        """Helper method to autocomplete a name and return matches

        Args:
            name (str): Name to autocomplete
            limit (int): Maximum number of matches to return
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.
            **kwargs:
            offset (int): The offset to start returning tags from.

        Returns:
            List: Autocomplete matches
        """
        hdxobject = cls(configuration=configuration)
        data = {"q": name, "limit": limit}
        data.update(kwargs)
        return hdxobject._write_to_hdx("autocomplete", data)

    def _addupdate_hdxobject(
        self,
        hdxobjects: List["HDXObject"],
        id_field: str,
        new_hdxobject: "HDXObject",
    ) -> "HDXObject":
        """Helper function to add a new HDX object to a supplied list of HDX objects or update existing metadata if the object
        already exists in the list

        Args:
            hdxobjects (List[HDXObject]): list of HDX objects to which to add new objects or update existing ones
            id_field (str): Field on which to match to determine if object already exists in list
            new_hdxobject (HDXObject): The HDX object to be added/updated

        Returns:
            HDXObject: The HDX object which was added or updated
        """
        for hdxobject in hdxobjects:
            if hdxobject[id_field] == new_hdxobject[id_field]:
                merge_two_dictionaries(hdxobject, new_hdxobject)
                return hdxobject
        hdxobjects.append(new_hdxobject)
        return new_hdxobject

    def _remove_hdxobject(
        self,
        objlist: List[Union["HDXObject", Dict]],
        obj: Union["HDXObject", Dict, str],
        matchon: str = "id",
        delete: bool = False,
    ) -> bool:
        """Remove an HDX object from a list within the parent HDX object

        Args:
            objlist (List[Union[HDXObject,Dict]]): list of HDX objects
            obj (Union[HDXObject,Dict,str]): Either an id or hdx object metadata either from an HDX object or a dictionary
            matchon (str): Field to match on. Defaults to id.
            delete (bool): Whether to delete HDX object. Defaults to False.

        Returns:
            bool: True if object removed, False if not
        """
        if objlist is None:
            return False
        if isinstance(obj, str):
            obj_id = obj
        elif isinstance(obj, dict) or isinstance(obj, HDXObject):
            obj_id = obj.get(matchon)
        else:
            raise HDXError("Type of object not a string, dict or T<=HDXObject")
        if not obj_id:
            return False
        for i, objdata in enumerate(objlist):
            objid = objdata.get(matchon)
            if objid and objid == obj_id:
                if delete:
                    objlist[i].delete_from_hdx()
                del objlist[i]
                return True
        return False

    def _convert_hdxobjects(self, hdxobjects: List["HDXObject"]) -> List[Dict]:
        """Helper function to convert supplied list of HDX objects to a list of dict

        Args:
            hdxobjects (List[HDXObject]): List of HDX objects to convert

        Returns:
            List[Dict]: List of HDX objects converted to simple dictionaries
        """
        newhdxobjects = list()
        for hdxobject in hdxobjects:
            newhdxobjects.append(hdxobject.data)
        return newhdxobjects

    def _copy_hdxobjects(
        self,
        hdxobjects: List["HDXObject"],
        hdxobjectclass: type,
        attribute_to_copy: Optional[str] = None,
    ) -> List["HDXObject"]:
        """Helper function to make a deep copy of a supplied list of HDX objects

        Args:
            hdxobjects (List[HDXObject]): list of HDX objects to copy
            hdxobjectclass (type): Type of the HDX Objects to be copied
            attribute_to_copy (Optional[str]): An attribute to copy over from the HDX object. Defaults to None.

        Returns:
            List[HDXObject]: Deep copy of list of HDX objects
        """
        newhdxobjects = list()
        for hdxobject in hdxobjects:
            newhdxobjectdata = copy.deepcopy(hdxobject.data)
            newhdxobject = hdxobjectclass(
                newhdxobjectdata, configuration=self.configuration
            )
            if attribute_to_copy:
                value = getattr(hdxobject, attribute_to_copy)
                setattr(newhdxobject, attribute_to_copy, value)
            newhdxobjects.append(newhdxobject)
        return newhdxobjects

    def _separate_hdxobjects(
        self,
        hdxobjects: List["HDXObject"],
        hdxobjects_name: str,
        id_field: str,
        hdxobjectclass: type,
    ) -> None:
        """Helper function to take a list of HDX objects contained in the internal dictionary and add them to a
        supplied list of HDX objects or update existing metadata if any objects already exist in the list. The list in
        the internal dictionary is then deleted.

        Args:
            hdxobjects (List[HDXObject]): list of HDX objects to which to add new objects or update existing ones
            hdxobjects_name (str): Name of key in internal dictionary from which to obtain list of HDX objects
            id_field (str): Field on which to match to determine if object already exists in list
            hdxobjectclass (type): Type of the HDX Object to be added/updated

        Returns:
            None
        """
        new_hdxobjects = self.data.get(hdxobjects_name, list())
        """:type : List["HDXObject"]"""
        if new_hdxobjects:
            hdxobject_names = set()
            for hdxobject in hdxobjects:
                hdxobject_name = hdxobject[id_field]
                hdxobject_names.add(hdxobject_name)
                for new_hdxobject in new_hdxobjects:
                    if hdxobject_name == new_hdxobject[id_field]:
                        merge_two_dictionaries(hdxobject, new_hdxobject)
                        break
            for new_hdxobject in new_hdxobjects:
                if new_hdxobject[id_field] not in hdxobject_names:
                    hdxobjects.append(
                        hdxobjectclass(
                            new_hdxobject, configuration=self.configuration
                        )
                    )
            del self.data[hdxobjects_name]

    def _get_tags(self) -> List[str]:
        """Return the dataset's list of tags

        Returns:
            List[str]: list of tags or [] if there are none
        """
        tags = self.data.get("tags", None)
        if not tags:
            return list()
        return [x["name"] for x in tags]

    def _add_tag(self, tag: str, vocabulary_id: Optional[str] = None) -> bool:
        """Add a tag

        Args:
            tag (str): Tag to add
            vocabulary_id (Optional[str]): Vocabulary tag is in. Defaults to None.

        Returns:
            bool: True if tag added or False if tag already present
        """
        tag = tag.lower()
        tags = self.data.get("tags", None)
        if tags:
            if tag in [x["name"] for x in tags]:
                return False
        else:
            tags = list()
        tagdict = {"name": tag}
        if vocabulary_id is not None:
            tagdict["vocabulary_id"] = vocabulary_id
        tags.append(tagdict)
        self.data["tags"] = tags
        return True

    def _add_tags(
        self, tags: List[str], vocabulary_id: Optional[str] = None
    ) -> List[str]:
        """Add a list of tag

        Args:
            tags (List[str]): list of tags to add
            vocabulary_id (Optional[str]): Vocabulary tag is in. Defaults to None.

        Returns:
            List[str]: Tags that were successfully added
        """
        added_tags = list()
        for tag in tags:
            if self._add_tag(tag, vocabulary_id=vocabulary_id):
                added_tags.append(tag)
        return added_tags

    def _get_stringlist_from_commastring(self, field: str) -> List[str]:
        """Return list of strings from comma separated list

        Args:
            field (str): Field containing comma separated list

        Returns:
            List[str]: List of strings
        """
        strings = self.data.get(field)
        if strings:
            return strings.split(",")
        else:
            return list()

    def _add_string_to_commastring(self, field: str, string: str) -> bool:
        """Add a string to a comma separated list of strings

        Args:
            field (str): Field containing comma separated list
            string (str): String to add

        Returns:
            bool: True if string added or False if string already present
        """
        if string in self._get_stringlist_from_commastring(field):
            return False
        strings = f"{self.data.get(field, '')},{string}"
        if strings[0] == ",":
            strings = strings[1:]
        self.data[field] = strings
        return True

    def _add_strings_to_commastring(
        self, field: str, strings: List[str]
    ) -> bool:
        """Add a list of strings to a comma separated list of strings

        Args:
            field (str): Field containing comma separated list
            strings (List[str]): list of strings to add

        Returns:
            bool: True if all strings added or False if any already present.
        """
        allstringsadded = True
        for string in strings:
            if not self._add_string_to_commastring(field, string):
                allstringsadded = False
        return allstringsadded

    def _remove_string_from_commastring(self, field: str, string: str) -> bool:
        """Remove a string from a comma separated list of strings

        Args:
            field (str): Field containing comma separated list
            string (str): String to remove

        Returns:
            bool: True if string removed or False if not
        """
        commastring = self.data.get(field, "")
        if string in commastring:
            self.data[field] = commastring.replace(string, "")
            return True
        return False



"""Helper to the Dataset and Resource classes for handling HDX dates.
"""
from collections.abc import Iterable
from datetime import date, datetime
from typing import Dict, List, Optional, Tuple, Union

from hdx.utilities.dateparse import parse_date


class DateHelper:
    @staticmethod
    def get_date_info(
        hdx_date: Dict,
        date_format: Optional[str] = None,
        today: date = datetime.now(),
    ) -> Dict:
        """Get date as datetimes and strings in specified format. If no format is supplied, the ISO 8601
        format is used. Returns a dictionary containing keys startdate (start date as datetime), enddate (end
        date as datetime), startdate_str (start date as string), enddate_str (end date as string) and ongoing
        (whether the end date is a rolls forward every day).

        Args:
            hdx_date (str): Input date
            date_format (Optional[str]): Date format. None is taken to be ISO 8601. Defaults to None.
            today (date): Date to use for today. Defaults to datetime.now.

        Returns:
            Dict: Dictionary of date information
        """
        result = dict()
        if hdx_date:
            if hdx_date[0] == "[" and hdx_date[-1] == "]":
                hdx_date = hdx_date[1:-1]
            dataset_dates = hdx_date.split(" TO ")
            startdate = parse_date(dataset_dates[0])
            enddate = dataset_dates[1]
            if enddate == "*":
                enddate = datetime(today.year, today.month, today.day, 0, 0)
                ongoing = True
            else:
                enddate = parse_date(enddate)
                ongoing = False
            result["startdate"] = startdate
            result["enddate"] = enddate
            if date_format is None:
                startdate_str = startdate.isoformat()
                enddate_str = enddate.isoformat()
            else:
                startdate_str = startdate.strftime(date_format)
                enddate_str = enddate.strftime(date_format)
            result["startdate_str"] = startdate_str
            result["enddate_str"] = enddate_str
            result["ongoing"] = ongoing
        return result

    @staticmethod
    def get_hdx_date(
        startdate: Union[datetime, str],
        enddate: Union[datetime, str, None] = None,
        ongoing: bool = False,
    ) -> str:
        """Get an HDX date from either datetime.datetime objects or strings with option to set ongoing.

        Args:
            startdate (Union[datetime, str]): Dataset start date
            enddate (Union[datetime, str, None]): Dataset end date. Defaults to None.
            ongoing (bool): True if ongoing, False if not. Defaults to False.

        Returns:
            str: HDX date
        """
        if isinstance(startdate, str):
            startdate = parse_date(startdate)
        startdate = startdate.isoformat()
        if ongoing:
            enddate = "*"
        elif not enddate:
            enddate = startdate
        if enddate != "*":
            if isinstance(enddate, str):
                enddate = parse_date(enddate)
            enddate = enddate.isoformat()
        return f"[{startdate} TO {enddate}]"

    @classmethod
    def get_hdx_date_from_years(
        cls,
        startyear: Union[str, int, Iterable],
        endyear: Union[str, int, None] = None,
    ) -> Tuple[str, List[int]]:
        """Get an HDX date from an iterable of years or a start and end year.

        Args:
            startyear (Union[str, int, Iterable]): Start year given as string or int or range in an iterable
            endyear (Union[str, int, None]): End year given as string or int

        Returns:
            Tuple[str,List[int]]: (HDX date, the start and end year if supplied or sorted list of years)
        """
        retval = list()
        if isinstance(startyear, str):
            startyear = int(startyear)
        if not isinstance(startyear, Iterable):
            retval.append(startyear)
            if endyear is not None:
                if isinstance(endyear, str):
                    endyear = int(endyear)
                retval.append(endyear)
        else:
            retval = sorted(int(x) for x in list(startyear))
        startyear = retval[0]
        endyear = retval[-1]
        startdate = datetime(startyear, 1, 1)
        enddate = datetime(endyear, 12, 31)
        return cls.get_hdx_date(startdate, enddate), retval
